/*
 * Copyright (c) 2019 Peter Lager
 * <quark(a)lagers.org.uk> http:www.lagers.org.uk
 * 
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it freely,
 * subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented;
 * you must not claim that you wrote the original software.
 * If you use this software in a product, an acknowledgment in the product
 * documentation would be appreciated but is not required.
 * 
 * 2. Altered source versions must be plainly marked as such,
 * and must not be misrepresented as being the original software.
 * 
 * 3. This notice may not be removed or altered from any source distribution.
 */

import processing.core.*;

/**
 * Creates a 2D non-rectangular quad that can be textured with an image and
 * avoid the diagonal distortion that can result. It does this by splitting
 * the quad into smaller quads.
 * 
 * The class uses parametric equations to control how much of the quad is 
 * visible see interpolate(float) method for more detail. It also provides
 * simple animation to control how much of the QG is visible by hooking into
 * Processing's animation thread. Alternatively the user can ignore this 
 * and perform the animation through their own code.
 * 
 * The user can specify the amount of zoom and centre of zoom to be used 
 * during animation. To facilitate using this QG with other visual elements
 * the user can get the current bounds of the quad.
 * 
 * The QuadGrid texture allows the user to specify the texture coordinates 
 * allowing full control over how the QG is textured including tiling and 
 * horizontal / vertical fliping. 
 * 
 * Instances of QuadGrid cannot be created directly instead it uses the 
 * 'builder' pattern because of the large number of variables used to control
 * its appearance.
 * 
 * @author Peter Lager (aka quark)
 *
 */
public class QuadGrid implements PConstants {

  private static final int SE0 = 0;
  private static final int SE1 = 1;
  private static final int ME0 = 2;
  private static final int ME1 = 3;

  // PApplet instance to enable animation
  PApplet pa = null;
  // Texture attributes
  private PImage img;
  private int LAST_ROW, LAST_COL;
  private VPoint[][] vp;
  private float[] u = new float[2];
  private float[] v = new float[2];
  // Quad bounds and zoom attributes
  private VPoint[] corner = new VPoint[4];
  private float zoomFactor = 1;
  private VPoint zoomOrigin = null;
  // Animation attributes
  //   Parametric variable to control current state of the animation
  private float start_t = 0;
  private float end_t = 1;
  private float current_t = 1;
  //   Time in milliseconds to control animation
  private long anim_start_time = 0;
  private long anim_duration = 0;
  private boolean is_animating = false;

  /**
   * Used internally to adjust a quad corner position based on 
   * zoom factor and zoom centre
   * @param p the point to be adjusted
   * @param zo the origin for the zoom
   * @param zf the zoom factor
   */
  private void relocate(VPoint p, VPoint zo, float zf) {
    p.x = zo.x + (p.x - zo.x) * zf;
    p.y = zo.y + (p.y - zo.y) * zf;
  }

  /**
   * NEVER call this method directly it is used internally
   * for animation. It has public access because Processing 
   * requires it.
   */
  public void pre() {
    long time = System.currentTimeMillis();
    float d = ((float)(time - anim_start_time)) / ((float)anim_duration);
    if (d > 1) {
      current_t = end_t;
      is_animating = false;
      pa.unregisterMethod("pre", this);
    } else {
      current_t = start_t + (end_t - start_t) * d;
    }
    interpolate(current_t);
  }

  /**
   * Request an animation of this QuadGrid instance. If the user does not provide
   * and instance of PApplet or is in a state of animation the request is denied
   * and the method returns false.
   * Both start_t and sned_t should be in the inclusive range 0 (hidden) and (1)
   * fully visible.
   *   
   * @param pa instance of PApplet to hook into
   * @param start_t starting value for the animation
   * @param end_t ending value for the animation
   * @param anim_duration the duration of the animation in milliseconds
   * @return true if the animation was started else false
   */
  public boolean anim(PApplet pa, float start_t, float end_t, long anim_duration) {
    if (pa == null || is_animating) {
      return false;
    }
    this.start_t = start_t;
    this.current_t = start_t;
    this.end_t = end_t;
    this.anim_duration = anim_duration;
    this.anim_start_time = System.currentTimeMillis();
    is_animating = true;
    this.pa = pa;
    this.pa.registerMethod("pre", this);
    return true;
  }

  /**
   * Request how much of the QuadGrid instance is to be displayed. If this
   * instance is being animated then the request is rejected and the method
   * returns false.
   * 
   * Use this method if you are providing your own animation code.
   * 
   * t should be in the inclusive range 0 (hidden) and (1) fully visible.
   * 
   * 
   * @param t parametric value in the range 0 - 1 inclusive
   * @return true if the request is implemented alse false
   */
  public boolean show(float t) {
    if (!is_animating) {
      interpolate(t);
      return true;
    }
    return false;
  }

  /**
   * Parametric method used internally to calculate both the XY and the
   * texture coordinates for ths QuadGrid instance.
   * 
   * t should be in the inclusive range 0 (hidden) and (1) fully visible.
   * 
   * @param t the 
   */
  private void interpolate(float t) {
    // Set the corners for static edge
    vp[0][0].setXY(corner[SE0]);
    vp[0][LAST_ROW].setXY(corner[SE1]);
    // Set the corners for moving edge
    vp[LAST_COL][0].pointXY(t, corner[SE0], corner[ME0]);
    vp[LAST_COL][LAST_ROW].pointXY(t, corner[SE1], corner[ME1]);
    // Adjust for zoom
    if (zoomFactor != 1) {
      float zf = zoomFactor + (1 - zoomFactor) * t;
      relocate(vp[0][0], zoomOrigin, zf);
      relocate(vp[0][LAST_ROW], zoomOrigin, zf);
      relocate(vp[LAST_COL][0], zoomOrigin, zf);
      relocate(vp[LAST_COL][LAST_ROW], zoomOrigin, zf);
    }
    // Set the static and moving edge XY coordinates
    float deltaEdgeT = 1.0f / LAST_ROW;
    for (int r = 1; r < LAST_ROW; r ++) {
      vp[0][r].pointXY(r * deltaEdgeT, vp[0][0], vp[0][LAST_ROW]);
      vp[LAST_COL][r].pointXY(r * deltaEdgeT, vp[LAST_COL][0], vp[LAST_COL][LAST_ROW]);
    }
    // Set XY for all points between static and moving edges
    float deltaT = 1.0f / LAST_COL;
    for (int c = 1; c < LAST_COL; c++) {
      for (int r = 0; r <= LAST_ROW; r ++) {
        vp[c][r].pointXY(c * deltaT, vp[0][r], vp[LAST_COL][r]);
      }
    }
    // Set UV for all points
    float deltaU =  (u[1] - u[0]) * t / LAST_COL;
    float deltaV = (v[1] - v[0]) / LAST_ROW;
    for (int c = 0; c <= LAST_COL; c++) {
      for (int r = 0; r <= LAST_ROW; r++) {
        vp[c][r].setUV(u[0] + deltaU * c, v[0] + deltaV * r);
      }
    }
  }

  /**
   * Returns true if it is being animated on the Processing thread
   * using the anim() method.
   */
  public boolean isAnimating() {
    return is_animating;
  }

  /**
   * Returns the current bounds of this QuadGrid instance.
   * The elements of the array are
   * [0] static edge top 
   * [1] static edge bottom
   * [1] moving edge top
   * [3] moving edge bottom
   */
  public PVector[] bounds() {
    return new PVector[] {
      vp[0][0].copyAsPVector(), 
      vp[0][LAST_ROW].copyAsPVector(), 
      vp[LAST_COL][0].copyAsPVector(), 
      vp[LAST_COL][LAST_ROW].copyAsPVector()
    };
  }

  /**
   * Displays the grid on the PApplet instance specified.
   */
  public void drawGrid(PApplet app) {
    if (vp == null) return;
    app.textureMode(PApplet.NORMAL);
    app.noStroke();
    app.textureWrap(REPEAT);
    //    app.stroke(255,0,255); // comment out this line to see triangles
    for (int row = 0; row < LAST_ROW; row++) {
      app.beginShape(PApplet.TRIANGLE_STRIP);
      app.texture(img);
      for (int col = 0; col <= LAST_COL; col++) {
        VPoint p0 = vp[col][row];
        VPoint p1 = vp[col][row+1];
        app.vertex(p0.x, p0.y, p0.u, p0.v);
        app.vertex(p1.x, p1.y, p1.u, p1.v);
      }
      app.endShape();
    }
  }

  /**
   * Private ctor - user must create a Builder instance to create QuadGrids.
   */
  private QuadGrid(PImage img, VPoint[] corner, int nbrCols, int nbrRows, 
    float[] u, float[] v, VPoint zoomOrigin, float zoomFactor) 
    throws IllegalStateException {
    this.img = img;
    if (this.img == null) {
      throw new IllegalStateException("Image undefined");
    } else {
      // Use a copy of the image to avoid memory leak when this object goes out of scope
      this.img = img.get();
    }
    // Make a copy of the corner array to avoid side effects if reusing the builder
    for (int i = 0; i < corner.length; i++) {
      this.corner[i] = corner[i] == null ? null : corner[i].copy();
    }
    if (this.corner[SE0] == null || this.corner[SE1] == null) {
      throw new IllegalStateException("Static edge undefined");
    }
    if (this.corner[ME0] == null || this.corner[ME1] == null) {
      throw new IllegalStateException("Moving edge undefined");
    }
    LAST_COL = nbrCols;
    LAST_ROW = nbrRows;
    vp = new VPoint[LAST_COL + 1][LAST_ROW + 1];
    for (int c = 0; c < vp.length; c++) {
      for (int r = 0; r < vp[c].length; r++) {
        vp[c][r] = new VPoint();
      }
    }
    // Copy the UV arrays to avoid side effects if reusing the builder
    for (int i = 0; i < 2; i++) {
      this.u[i] = u[i];
      this.v[i] = v[i];
    }
    this.zoomFactor = zoomFactor;
    this.zoomOrigin = zoomOrigin == null ? null : zoomOrigin.copy();
    if (this.zoomFactor != 1) {
      if (this.zoomOrigin == null) {
        throw new IllegalStateException("Zoom origin undefined");
      }
    }
  }

  /**
   * A simple class used internally by QuadGrid to hold the XY and texture
   * coordinates of a point in 2D space. 
   * 
   * @author Peter Lager
   *
   */
  protected static class VPoint {
    protected float x = 0;
    protected float y = 0;
    protected float u = 0;
    protected float v = 0;

    public VPoint() {
    }

    public VPoint(float x, float y) {
      this.x = x;
      this.y = y;
    }

    public VPoint(VPoint p ) {
      this.x = p.x;
      this.y = p.y;
      this.u = p.u;
      this.v = p.v;
    }

    public VPoint copy() {
      return new VPoint(this);
    }

    public PVector copyAsPVector() {
      return new PVector(x, y);
    }

    public void setUV(float u, float v) {
      this.u = u;
      this.v = v;
    }

    public void setXY(float x, float y) {
      this.x = x;
      this.y = y;
    }

    public void setXY(VPoint p) {
      this.x = p.x;
      this.y = p.y;
    }

    public void pointXY(float t, VPoint p0, VPoint p1) {
      float nx = p0.x + (p1.x - p0.x) * t;
      float ny = p0.y + (p1.y - p0.y) * t;
      x = nx;
      y = ny;
    }
  }

  /**
   * This is the Builder class used to create QuadGrid instances.
   * As ingle builder object can create multiple instances of 
   * QuadGrid that vary slightly.
   * 
   * @author Peter Lager
   */
  public static class Builder {
    private PImage img = null;
    private int nbrCols = 1, nbrRows = 1;

    private float[] u = { 0, 1 };
    private float[] v = { 0, 1 };

    private  VPoint[] corner =  new VPoint[4];

    private VPoint zoomOrigin = null;
    private float zoomFactor = 1.0f;

    /**
     * Specifies the image to be used as the texture
     * @param img the texture image
     */
    public Builder useImage(PImage img) {
      this.img = img != null ? img : this.img;
      return this;
    }

    /**
     * Defines the number of quad tiles to use. Both parameter 
     * values will be forced into the range 1-100 inclusive.
     * @param nbrCols number of horizontal quads
     * @param nbrRows number of vertical quads
     */
    public Builder detail(int nbrCols, int nbrRows) {
      this.nbrCols = nbrCols > 0 && nbrCols <= 100 ? nbrCols : 1;
      this.nbrRows = nbrRows > 0 && nbrRows <= 100 ? nbrRows : 1;
      return this;
    }

    /**
     * Defines the texture coordinates to be used on the QuadGrid
     * instance.
     */
    public Builder uvCoords(float u0, float u1, float v0, float v1) {
      this.u[0] = u0;
      this.u[1] = u1;
      this.v[0] = v0;
      this.v[1] = v1;
      return this;
    }

    /**
     * The top and bottom display coordinates for the static edge.
     * This equates to the left hand side of the image when using
     * the default texture coordinates.
     */
    public Builder staticEdge(float xTop, float yTop, float xBot, float yBot) {
      corner[SE0] = new VPoint(xTop, yTop);
      corner[SE1] = new VPoint(xBot, yBot);
      return this;
    }

    /**
     * The top and bottom display coordinates for the moving edge.
     * This equates to the right hand side of the image when using
     * the default texture coordinates.
     */
    public Builder movingEdge(float xTop, float yTop, float xBot, float yBot) {
      corner[ME0] = new VPoint(xTop, yTop);
      corner[ME1] = new VPoint(xBot, yBot);
      return this;
    }

    /**
     * Defines the centre coordinates for the zoom. The zoom factor
     * is 1.0 when the QG is fully visible, in this method the last
     * parameter is the zoom to use when the QG first appears.
     * 
     * @param x horizontal coordinate for zoom centre
     * @param y vertical coordinate for zoom centre
     * @param zf the zoom factor when t = 0
     */
    public Builder zoom(float x, float y, float zf) {
      zoomOrigin = new VPoint( x, y);
      zoomFactor = zf;
      return this;
    }

    /**
     * If possible build an instance of QuadGrid based on values stored
     * in this builder. If an error occurs then a message will appear 
     * in the error / output console and the method returns null.
     * 
     * @return a QuadGrid instance or null if the build fails.
     */
    public QuadGrid build() {
      QuadGrid qg = null;
      try {
        qg = new QuadGrid(img, corner, nbrCols, nbrRows, u, v, zoomOrigin, zoomFactor);
      }
      catch (IllegalStateException e) {
        System.err.println("QuadGrid builder error : " + e.getMessage());
      }
      return qg;
    }
  }
}
